//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	tcsfix	--- Time Correlation Spacecraft file FIXing task
//
//	DESCRIPTION:
//		This task fixes two distinct problems in ODF time correlation data:
//
//		1. Due to wrong GS+SC time propagation delays hardcoded in XSCS<=v71
//		   (see email Ed Serpel, dated 2001-02-22) all time correlation files
//		   produced with these XSCS versions contains too small UTC values.
//		   The absolute error is - depending slighlty on the GS ID - of the
//		   order of 60ms. In addition some of the early TCS files do not
//		   contain the ERTIMRAW column that hold the non-propagation delay
//		   corrected earth reception times. Also, some early TCS have bogus
//		   entries with UTCDAY=0 - those rows are removed from the table.
//
//		2. During revolutions 384 - 394 (both inclusive) the Santiago
//		   ground station clock was running exactly one second ahead of
//		   UTC (see emails Ed Serpell, dated 2002-02-26 + Ramon Munoz
//		   dated 2002-02-27 for a list of affected observations). Consequently,
//		   the time correlation data show a corresponding discontinuity at
//		   the time of a station handover to Santiago (GSID==41).
//
//		tcsfix does:
//			* remove rows where UTCDAY=0
//			* add the ERTTIMRAW column to the TCS unless it exists already
//		C1	if (rev < 207)
//				* undo the wrong propagation delays to all UTC fields
//				  and applies the correct delays
//		C2	if (rev >= 384 && rev <= 394)
//				* subtract 1s to UTC data from Santiago station
//			* adds attribues
//				TCSFIXED = T        // this TCS has been fixed
//				TCSFIXV  = M.m.p	// tcsfix version number
//
//		C1 is only done if the revolution number is less than 207
//		(XSCS-v72 was put online at the beginning of rev 207)
//		C2 is only done for ODFs from revolutions [384, 394]
//		C1/C2 are not performed if the attribute TCSFIXED is set to 'T' to
//		avoid double-correction throuh repeated invocations of tcsfix.
//
//	USAGE:
//		tcsfix [clonetcs=yes|no] [rerun=yes|no]
//
//		By default tcsfix replaces the original TCS data (in the ODF pointed
//		at via environment variable SAS_ODF) with the corrected ones.
//		clonetcs=yes forces a copy of the orignal data file as '...TCS.FIT.old'
//		to be created prior to processing. Unless rerun=yes, tcsfix refuses
//		to run again on a TCS which has been corrected already
//
//	CHANGE HISTORY:
//		UL, 2001-03-27: 1.0.0: - initial version
//		UL, 2001-05-31: 1.1.0: - fixed following problems (found by running
//								 on ODF 0078_0124100101):
//									+ put try-catch block around
//										posloc.stateVector(t)
//									  to catch exceptions thrown when there
//									  is no orbit data for particular
//									  TCS entries; in this case the
//									  ERTIMRAW time is set to
//									  Xmm::missionReferenceTime
//									+ ignore bogus TCS entries with UTCDAY==0
//									  (TCSs in early ODF seem to be afflicted
//									   with this problem)
//		UL, 2002-04-08: 1.2.0: - extended functionality to also correct for
//								 Santiago-1s problem
//
//	AUTHOR:
//		U. Lammers, Tue Mar 27 15:28:54 MET DST 2001
//
//	$Header: /usr/local/cvs/sasdev/task_devel/packages/tools/src/tcsfix/tcsfix.cc,v 1.5 2002/04/25 13:34:41 ulammers Exp $
//

#include <cmath>

#include "DataFilter.h"
#include "OalServer.h"
#include "XmmPositionLocator.h"
#include "Xmm.h"
#include "PhysicalConstants.h"
#include "CalOalUtility.h"
#include "STime.h"

#include "tcsfix.h"

class GSTime {
public:
	GSTime(int16 day, int32 milli, int16 micro) :
			_day(day), _milli(milli), _micro(micro) {}

	int16 day() const { return _day; }
	int32 milli() const { return _milli; }
	int16 micro() const { return _micro; }

	GSTime &operator-=(double delta) {
		assert(delta>=0.);

		unsigned udelta = UNT(1000*1000*delta+.5);

		unsigned microsec = udelta%1000;
		udelta /= 1000;
		unsigned millisec = udelta%86400000;

		_micro -= microsec;
		_milli -= millisec;
		if (_micro<0) {
			_micro += 1000;
			--_milli;
		}
		if (_milli<0) {
			_milli += 1000*86400;
			--_day;
			assert(_day>=0);	//	cannot get negative
		}

		return *this;
	} // operator-=

	GSTime &operator+=(double delta) {
		assert(delta>=0.);

		unsigned udelta = UNT(1000*1000*delta+.5);

		unsigned microsec = udelta%1000;
		udelta /= 1000;
		unsigned millisec = udelta%86400000;

		_micro += microsec;
		_milli += millisec;
		if (_micro>=1000) {
			_micro -= 1000;
			++_milli;
		}
		if (_milli>=1000*86400) {
			_milli -= 1000*86400;
			++_day;
		}

		return *this;
	} // operator+=

private:
	int16	_day;
	int32	_milli;
	int16	_micro;
};

class TcsFixer {
public:
	TcsFixer(Table *tcstab) : _tcstab(tcstab) {}

	void correctSigDel(const Odf *odf) {
		const double GS_DELAY_WRONG = 33896.e-6,	//	Constant wrong delays
				     SC_DELAY_WRONG = 33896.e-6;	//	in [s]; see email
													//	Ed Serpel dated
													//	2001-02-22
		const unsigned DELAY_INCORRECT_TILL_REV = 206;
													//	last revolution with
													//	incorrect TCS

		if (odf->orbitNr() > DELAY_INCORRECT_TILL_REV) {
			errstr<<"This ODF is from revolution "<<odf->orbitNr()<<" - as of "
					"revolution "<<DELAY_INCORRECT_TILL_REV+1
				  <<" the GS+SC time propagation delays are correct - "
					"skipping this stage ..."
				  <<Sas::message(Msg::App, Msg::Always);
			return;
		}

		clean();

		unsigned N = _tcstab->rows();
		ColumnData *erttimecol = _tcstab->column("ERTIME")->data(),
				   *cdutcday = _tcstab->column("UTCDAY")->data(),
				   *cdutcmilli = _tcstab->column("UTCMILLI")->data(),
				   *cdutcmicro = _tcstab->column("UTCMICRO")->data(),
				   *erttimeraw = 0;
		if (_tcstab->hasColumn("ERTIMRAW"))
			erttimeraw = _tcstab->column("ERTIMRAW")->data();
		else {
			//	old-style - create it first
			longVec lv(1); lv[0] = 27;
			erttimeraw = _tcstab->addColumn("ERTIMRAW", Dal::DString,
								"raw earth reception time", "", lv)->data();
		}

		int16	*utcday		= cdutcday->int16Data();
		int32	*utcmilli	= cdutcmilli->int32Data();
		int16	*utcmicro	= cdutcmicro->int16Data();

		XmmPositionLocator posloc(odf->selectSet(Xmm::SC,
											 Xmm::RECONSTRUCTED_ORBIT, 1));

		vector<string> gsid;
		CalOalUtility::fillVector(_tcstab, "GSID", gsid);
		const STime t1958(1958, 1, 1);
		for (unsigned n=0; n<N; ++n) {
			STime t(1958, 1, 1+utcday[n], 0, 0, 0,
					1.e-3 * DBL(utcmilli[n]) + 1.e-6 * DBL(utcmicro[n]));
			unsigned short gsi = toNumber<unsigned short>(gsid[n]);
			const XmmGroundStation &gs = XmmMission::groundStation(gsi);

			//	undo incorrect and apply correct delays
			t += GS_DELAY_WRONG + SC_DELAY_WRONG -
					XmmMission::dataHandler().timingDelay() - gs.timingDelay();

			double deltaT = t - t1958 - t.leapSeconds(),
				   days;
			double secs = STime::secondsPerDay*modf(
										deltaT/STime::secondsPerDay, &days);
			double msecs;
			double usecs = 1000.*modf(secs*1000., &msecs);

			utcday[n] = int16(days);
			utcmilli[n] = int32(msecs);
			utcmicro[n] = int16(usecs);

			CalOalUtility::stringColumnRowSet(erttimecol, n, t.toString(6));

			//	compute raw earth reception time
			try {
				CartesianVector xmmPos = posloc.stateVector(t).position,
							    gsPos = gs.stateVector(t).position;
				double xmmGSdist = (gsPos - xmmPos).norm() * 1000.,	//	km -> m
				propDelay =
						//	OBDH delay
						XmmMission::dataHandler().timingDelay() +

						//	signal travel time - NOTE: atmospheric propagation
						//	delays are of the order of 0.1 us (see email
						//	F. Dreger dated 2001-02-28), so, negligible
					    xmmGSdist/PhysicalConstants::c +

						// Ground station delays
						gs.timingDelay();
				t += propDelay;
			} catch(XmmPositionLocator::NoPositionDataAvailable) {
				t = Xmm::missionReferenceTime;
			}
			CalOalUtility::stringColumnRowSet(erttimeraw, n, t.toString(6));
		}

		delete erttimecol;
		delete cdutcday;
		delete cdutcmilli;
		delete cdutcmicro;
		delete erttimeraw;

		errstr<<"GS+SC time propagation delays successfully corrected in `"
			  <<_tcstab->qualifiedName()<<"'"
			  <<Sas::message(Msg::App, Msg::Always);
	} // correctSigDel()


	void correctSantiago(const Odf *odf) {
		const unsigned SANTIAGO_OFF_FROM_REV = 384,	//	first/last revolution
					   SANTIAGO_OFF_TILL_REV = 394;	//	in which Santiago GS
													//  clock is ahead of UTC
		const double   SANTIAGO_DELTA		 = 1.;	//	amount by which
													//	Santiago GS	clock is
													//	ahead of UTC

		if (odf->orbitNr() < SANTIAGO_OFF_FROM_REV ||
			odf->orbitNr() > SANTIAGO_OFF_TILL_REV) {
			errstr<<"This ODF is from revolution "<<odf->orbitNr()
				  <<" - Santiago GS clock was behind UTC by "
				  <<SANTIAGO_DELTA<<" during revolutions ["
				  <<SANTIAGO_OFF_FROM_REV<<", "<<SANTIAGO_OFF_TILL_REV
				  <<"] only - skipping this stage ..."
				  <<Sas::message(Msg::App, Msg::Always);
			return;
		}

		clean();

		vector<string> gsid;
		CalOalUtility::fillVector(_tcstab, "GSID", gsid);
		unsigned N = _tcstab->rows();

		ColumnData *cdutcday = _tcstab->column("UTCDAY")->data(),
				   *cdutcmilli = _tcstab->column("UTCMILLI")->data(),
				   *cdutcmicro = _tcstab->column("UTCMICRO")->data(),
				   *cdert = _tcstab->column("ERTIME")->data(),
				   *cdertraw = _tcstab->column("ERTIMRAW")->data();
		int16	*utcday		= cdutcday->int16Data();
		int32	*utcmilli	= cdutcmilli->int32Data();
		int16	*utcmicro	= cdutcmicro->int16Data();

		unsigned short SantiagoId = XmmMission::groundStation(
												XmmMission::Santiago).id();
		for (unsigned n=0; n<N; ++n)
			if (toNumber<unsigned short>(gsid[n]) == SantiagoId) {
				GSTime gst(utcday[n], utcmilli[n], utcmicro[n]);
				gst += SANTIAGO_DELTA;
				utcday[n] = gst.day();
				utcmilli[n] = gst.milli();
				utcmicro[n] = gst.micro();

				STime ert(CalOalUtility::stringColumnRow(cdert, n)),
					  ertraw(CalOalUtility::stringColumnRow(cdertraw, n));
				ert += SANTIAGO_DELTA;
				ertraw += SANTIAGO_DELTA;
				CalOalUtility::stringColumnRowSet(cdert, n, ert.toString(6));
				CalOalUtility::stringColumnRowSet(cdertraw, n,
														ertraw.toString(6));
			}

		delete cdutcday;
		delete cdutcmilli;
		delete cdutcmicro;

		errstr<<"Wrong Santiago GS clock values successfully corrected in `"
			  <<_tcstab->qualifiedName()<<"'"
			  <<Sas::message(Msg::App, Msg::Always);
	} // correctSantiago()


	//
	//	remove rows with UTCDAY==0
	//	occasionally seen in the early ODFs (e.g. rev78)
	//
	void clean() {
		DataFilter filter("UTCDAY!=0");
		filter.doFilter(*_tcstab);
	} // clean()

	~TcsFixer() {}

private:
	Table *_tcstab;
};

void tcsfix()
{
	const char *version = "1.1.0";				//	version

	//	locate TCS
	const Odf *odf = oalServer.odf();
	string tcsname = odf->selectSet(Xmm::SC, Xmm::TIME_CORRELATION, 1)->name();
	Table *tcstab = dataSetServer->open(tcsname, Dal::Modify)->table(0);
	if (tcstab->hasAttribute("TCSFIXED")) {
		errstr<<"`"<<tcstab->qualifiedName()<<"' has already been fixed "
				"in a previous run - will not do it again."
			  <<Sas::message(Msg::App, Msg::Always);
		dataSetServer->close(tcstab->parent());
		return;
	}

	if (booleanParameter("clonetcs"))
		dataSetServer->copy(tcsname, tcsname + ".old");

	TcsFixer tcsfixer(tcstab);
	tcsfixer.correctSigDel(odf);
	tcsfixer.correctSantiago(odf);

	//	add TCSFIXED
	*tcstab->addAttribute("TCSFIXED", "TCS has been fixed by tcsfix") = true;
	*tcstab->addAttribute("TCSFIXV", "tcsfix version that fixed this TCS") =
									version;
	dataSetServer->close(tcstab->parent());
}
